(* XML for the Interface between the isac-kernel and the java-frontend,
   authors: Walther Neuper 2002
   (c) due to copyright terms

This code remains in order to
keep tests of interaction Java-frontend <-> Kernel running.
*)

fun adduserOK2xml (calcid : States.calcID) (userid : States.iterID) =
  XML.Elem (("ADDUSER", []),
    [XML.Elem (("CALCID", []), [XML.Text (string_of_int calcid)]),
    XML.Elem (("USERID", []), [XML.Text (string_of_int userid)])])

fun calctreeOK2xml (calcid : States.calcID) = 
  XML.Elem (("CALCTREE", []),
    [XML.Elem (("CALCID", []), [XML.Text (string_of_int calcid)])])    
fun deconstructcalctreeOK2xml (calcid : States.calcID) = 
  XML.Elem (("DELCALC", []),
    [XML.Elem (("CALCID", []), [XML.Text (string_of_int calcid)])])    

fun iteratorOK2xml (calcid : States.calcID) (p : Pos.pos')= 
  XML.Elem (("CALCITERATOR", []),
    [XML.Elem (("CALCID", []), [XML.Text (string_of_int calcid)]),
    xml_of_pos "POSITION" p])
fun iteratorERROR2xml  (calcid : States.calcID) = 
  XML.Elem (("CALCITERATOR", []),
    [XML.Elem (("CALCID", []), [XML.Text (string_of_int calcid)]),
    XML.Elem (("ERROR", []), [XML.Text " iteratorERROR2xml: pos does not exist "])])

fun sysERROR2xml (calcid : States.calcID) str =
  XML.Elem (("SYSERROR", []),
    [XML.Elem (("CALCID", []), [XML.Text (string_of_int calcid)]),
    XML.Elem (("ERROR", []), [XML.Text (if str = "" then " ERROR in kernel " else str)])])

fun refformulaOK2xml (calcid : States.calcID) p (Ctree.Form t) = 
      XML.Elem (("REFFORMULA", []),
        [XML.Elem (("CALCID", []), [XML.Text (string_of_int calcid)]),
        XML.Elem (("CALCFORMULA", []), [
          xml_of_pos "POSITION" p,
          xml_of_term_NEW t])])
  | refformulaOK2xml (calcid : States.calcID) p (Ctree.ModSpec modspec) = 
      XML.Elem (("REFFORMULA", []), [
        XML.Elem (("CALCID", []), [XML.Text (string_of_int calcid)]),
        (*L.Elem (("CALCHEAD*) xml_of_posmodspec (p, modspec)])

fun gettacticOK2xml (calcid : States.calcID) tac = 
  XML.Elem (("GETTACTIC", []),[
    XML.Elem (("CALCID", []), [XML.Text (string_of_int calcid)]),
    xml_of_tac tac])

fun gettacticERROR2xml (calcid : States.calcID) str = 
  XML.Elem (("GETTACTIC", []),[
    XML.Elem (("CALCID", []), [XML.Text (string_of_int calcid)]),
    XML.Elem (("ERROR", []), [XML.Text  str])])

fun applicabletacticsOK calcid tacs =
  XML.Elem (("APPLICABLETACTICS", []), [
    XML.Elem (("CALCID", []), [XML.Text (string_of_int calcid)]),
    XML.Elem (("TACLIST", []), (map xml_of_tac tacs))])

fun getasmsOK2xml (calcid : States.calcID) terms = 
  XML.Elem (("ASSUMPTIONS", []), [
    XML.Elem (("CALCID", []), [XML.Text (string_of_int calcid)]),
    XML.Elem (("ASMLIST", []), (map xml_of_term_NEW terms))])

fun getintervalOK (calcid : States.calcID) fs =
      XML.Elem (("GETELEMENTSFROMTO", []),
        [XML.Elem (("CALCID", []), [XML.Text (string_of_int calcid)]),
        XML.Elem (("FORMHEADS", []), map xml_of_posterm fs)])

fun appendformulaOK2xml (calcid : States.calcID) (old : Pos.pos') (del : Pos.pos') (new : Pos.pos') =
  xml_of_calcchanged calcid "APPENDFORMULA" old del new
fun appendformulaERROR2xml (calcid : States.calcID) e =
  XML.Elem (("APPENDFORMULA", []), [
    XML.Elem (("CALCID", []), [XML.Text (string_of_int calcid)]),
    XML.Elem (("CALCMESSAGE", []), [XML.Text e])])

fun replaceformulaOK2xml (calcid : States.calcID) (old : Pos.pos') (del : Pos.pos') (new : Pos.pos') = 
  xml_of_calcchanged calcid "REPLACEFORMULA" old del new
fun replaceformulaERROR2xml (calcid : States.calcID) e =
  XML.Elem (("REPLACEFORMULA", []), [
    XML.Elem (("CALCID", []), [XML.Text (string_of_int calcid)]),
    XML.Elem (("CALCMESSAGE", []), [XML.Text e])])

fun message2xml (calcid : States.calcID) e =
  XML.Elem (("MESSAGE", []), [
    XML.Elem (("CALCID", []), [XML.Text (string_of_int calcid)]),
    XML.Elem (("STRING", []), [XML.Text e])])

fun setnexttactic2xml (calcid : States.calcID) e =
  XML.Elem (("SETNEXTTACTIC", []), [
    XML.Elem (("CALCID", []), [XML.Text (string_of_int calcid)]),
    XML.Elem (("MESSAGE", []), [XML.Text e])])

fun fetchproposedtacticOK2xml (calcid : States.calcID) tac (errpatIDs : Error_Pattern_Def.id list) =
  XML.Elem (("NEXTTAC", []), [
    XML.Elem (("CALCID", []), [XML.Text (string_of_int calcid)]),
    XML.Elem (("TACTICERRORPATTERNS", []), [xml_of_strs errpatIDs]),
    xml_of_tac tac])

fun fetchproposedtacticERROR2xml (calcid : States.calcID) e =
  XML.Elem (("NEXTTAC", []), [
    XML.Elem (("CALCID", []), [XML.Text (string_of_int calcid)]),
    XML.Elem (("ERROR", []), [XML.Text e])])

fun autocalculateOK2xml (calcid : States.calcID) (old : Pos.pos') (del : Pos.pos') (new : Pos.pos') = 
  xml_of_calcchanged calcid "AUTOCALC" old del new
fun autocalculateERROR2xml (calcid : States.calcID) e =
  XML.Elem (("AUTOCALC", []), [
    XML.Elem (("CALCID", []), [XML.Text (string_of_int calcid)]),
    XML.Elem (("CALCMESSAGE", []), [XML.Text e])])

fun interStepsOK (calcid : States.calcID) (old : Pos.pos') (del : Pos.pos') (new : Pos.pos') =
   xml_of_calcchanged calcid "INTERSTEPS" old del new
fun interStepsERROR (calcid : States.calcID) e =
  XML.Elem (("INTERSTEPS", []), [
    XML.Elem (("CALCID", []), [XML.Text (string_of_int calcid)]),
    XML.Elem (("CALCMESSAGE", []), [XML.Text e])])

fun calcMessage2xml (cI: States.calcID) e =
    writeln ("@@@@@begin@@@@@\n "^string_of_int cI^" \n" ^
	     "  <CALCMESSAGE> "^ e ^" </CALCMESSAGE>\n" ^
	     "@@@@@end@@@@@");

fun modifycalcheadOK2xml (calcid : States.calcID) (chd as (complete, _ ,_ ,_ ,_ ,_) : SpecificationC.T) =
  XML.Elem (("MODIFYCALCHEAD", []), [
    XML.Elem (("CALCID", []), [XML.Text (string_of_int calcid)]),
    XML.Elem (("STATUS", []), [
      XML.Text (if complete then "complete" else "incomplete")]),
    xml_of_modspec chd])

fun contextpblOK2xml calcid contpbl =
  XML.Elem (("CONTEXTPBL", []), [
    XML.Elem (("CALCID", []), [XML.Text (string_of_int calcid)]),
    xml_of_matchpbl contpbl])
fun contextmetOK2xml calcid contmet =
  XML.Elem (("CONTEXTMET", []), [
    XML.Elem (("CALCID", []), [XML.Text (string_of_int calcid)]),
    xml_of_matchmet contmet])

fun findFillpatterns2xml (calcid : States.calcID) pattIDs =
  XML.Elem (("FINDFILLPATTERNS", []), [
    XML.Elem (("CALCID", []), [XML.Text (string_of_int calcid)]),
    xml_of_strs pattIDs])

