(* states for calculation in global refs
   use"../states.sml";
   use"states.sml";
   *)

signature STATES =
sig
  eqtype iterID
  eqtype calcID
  val add_calc: Calc.state_pre -> calcID
  val get_calc: calcID -> Calc.state_pre
  val get_pos: calcID -> iterID -> Pos.pos'
  val del_calc: calcID -> calcID
  val add_user: calcID -> iterID
  val del_user: calcID -> iterID -> iterID
  val reset: unit -> unit
(*val reset_states: unit -> unit*)
  val upd_calc: calcID -> Calc.state_pre -> unit
  val upd_ipos: calcID -> iterID -> Pos.pos' -> unit
  val upd_tacis: calcID -> State_Steps.T -> unit

\<^isac_test>\<open>
(** )val states: (calcID * (Calc.state_pre * (iterID * Pos.pos') list)) list Synchronized.var( **)

(** )val del_assoc: (''a * 'b) list * ''a -> (''a * 'b) list( **)
(** )val del_assoc2: calcID -> iterID -> (calcID * ('a * (iterID * 'b) list)) list -> (
      calcID * ('a * (iterID * 'b) list)) list( **)
(** )val get_iterID: calcID -> (calcID * (Calc.state_pre * (iterID * Pos.pos') list)) list -> iterID( **)
(** )val new_key: (iterID * 'a) list -> iterID -> iterID( **)
(** )val overwrite2: (calcID * ('a * (iterID * 'b) list)) list * ((calcID * iterID) * 'b) -> 
      (calcID * ('a * (iterID * 'b) list)) list( **)
\<close>
  end

(**)
structure States(**): STATES(**) =
struct
(**)

type iterID = int;
type calcID = int;

(* holds calculations; these are read/updated from the java-frontend at each interaction*)
val states = Synchronized.var "isac_states" ([] :
  (calcID *      (* the id unique for a calculation                               *)
    (Calc.state_pre * (* the interpreter state                                         *) 
      (iterID *  (* 1 sets the 'active formula': a calc. can have several visitors*)
        Pos.pos'     (* for iterator of a user                                        *)
      (* TODO iterID * pos' should go to java-frontend                                  *)
  ) list)) list);

fun reset () = Synchronized.change states (fn _ => []);
(*
states:= [(3,(Calc.state_empty_pre, [(1,e_pos'),
			    (3,e_pos')])),
	  (4,(Calc.state_empty_pre, [(1,e_pos'),
			    (2,e_pos')]))];
*)

(** create new instances of users and ptrees
   new keys are the lowest possible in the association list **)

(* add users *)
fun new_key u n = case assoc (u, n) of 
  NONE => n 
| SOME _ => new_key u (n+1);
(*///10.10
fun get_calcID (u:(calcID * (Calc.state_pre * (iterID * pos') list)) list) = 
    (new_key u 1):calcID;*)
(*
val new_iterID = get_calcID (!states);
val it = 1 : int
states:= (!states) @ [(new_iterID, [])];
!states;
val it = [(3,[(#,#),(#,#)]),(4,[(#,#),(#,#)]),(1,[])]
*)

(*///7.10.03/// add states to a users active states
fun get_calcID (uI:iterID) (p:(iterID * (calcID * state) list) list) = 
  case assoc (p, uI) of 
    NONE => raise ERROR ("get_calcID: no iterID " ^ 
			  (string_of_int uI))
  | SOME ps => (new_key ps 1):calcID;
> get_calcID 1 (!states);  
val it = 1 : calcID
*)

(** retrieve, update, delete a state by iterID, calcID **)

(*///7.10.
fun get_state (uI:iterID) (pI:calcID) = get_cal uI pI (!states);
fun get_calc  (uI:iterID) (pI:calcID) = (snd o (get_cal uI pI)) (!states);
*)
fun get_calc  (cI: calcID) = 
  case assoc (Synchronized.value states, cI) of 
	  NONE => raise ERROR ("get_calc " ^ string_of_int cI ^ " not existent")
  | SOME (c, _) => 
    let
      (*val (_, tacis) = c
      val _ = if length tacis > 1 then raise ERROR "length tacis > 1" else ()
        in Test_Isac.thy raise ERROR in Interpret/inform.sml*)
    in c end;
fun get_pos (cI: calcID) (uI: iterID) = 
    case assoc (Synchronized.value states, cI) of 
	NONE => raise ERROR ("get_pos: calc " ^ (string_of_int cI) 
			     ^ " not existent")
      | SOME (_, us) => 
	(case assoc (us, uI) of 
	    NONE => raise ERROR ("get_pos: user " ^ (string_of_int uI) 
				 ^ " not existent")
	  | SOME p => p);


fun del_assoc ([],_) = []
  | del_assoc a =
  let fun del ([], _) ps = ps
	| del ((keyi, xi) :: pairs, key) ps =
    if key = keyi then ps @ pairs
    else del (pairs, key) (ps @ [(keyi, xi)])
  in del a [] end;
(*
> val ps =  [(1,"1"),(2,"2"),(3,"3"),(4,"4")];     
> del_assoc (ps,3);
val it = [(1,"1"),(2,"2"),(4,"4")] : (int * string) list
*)

(* delete doesn't report non existing elements *)
(*/////7.10.
fun del_assoc2 (uI:iterID) (pI:calcID) ps =
  let val new_ps = del_assoc (the (assoc (ps, uI)), pI)
  in overwrite (ps, (uI, new_ps)) end;*)
(*
> states:= del_assoc2 4 41 (!states);
> !states;
val it = [(3,[(#,#),(#,#),(#,#)]),(4,[(#,#)]),(1,[(#,#)])] : states

> del_user 3;
> !states;
val it = [(4,[(#,#)]),(1,[(#,#)])] : states
*)
fun del_assoc2 (cI: calcID) (uI: iterID) ps =
    case assoc (ps, cI) of
	NONE => ps
      | SOME (cs, us) => 
	overwrite (ps, (cI, (cs, del_assoc (us, uI))));
(*
> del_assoc2 4 1 (!states);
val it =
   [(3, (((EmptyPtree, ([], Und)), []), [(1, ([], Und)), (3, ([], Und))])),
    (4, (((EmptyPtree, ([], Und)), []), [(2, ([], Und))]))]*)

fun overwrite2 (ps, (((cI: calcID), (uI: iterID)), p)) =
    case assoc (ps, cI) of
	NONE => 
	raise ERROR ("overwrite2: calc " ^ (string_of_int uI) ^" not existent")
      | SOME (cs, us) =>
	overwrite (ps, (cI ,(cs, overwrite (us, (uI, p)))));

(* here is _no_ critical section on states; sufficient would be
  fun upd_calc cI cs =
    case assoc (Synchronized.value states, cI) of
      NONE => raise ERROR ...
    | SOME (_, us) => 
      Synchronized.change states (fn s => overwrite ...)
*)
fun upd_calc (cI: calcID) cs = Synchronized.change states
        (fn s => case assoc (s, cI) of 
	          NONE => raise ERROR ("upd_calc " ^ (string_of_int cI) ^ " not existent")
          | SOME (_, us) => overwrite (s, (cI, (cs, us))));
(*WN051210 testing before initac: only 1 taci in Calc.state_pre so far:
fun upd_calc (cI:calcID) (cs as (_, tacis):Calc.state_pre) =
    (if length tacis > 1 
     then raise ERROR ("upd_calc, |tacis|>1: "^tacis2str tacis) 
     else ();
    case assoc (!states, cI) of 
	NONE => raise ERROR ("upd_calc "^(string_of_int cI)^" not existent")
      | SOME (_, us) => states:= overwrite (!states, (cI, (cs, us)))
			);*)


(*///7.10.
fun upd_tacis (uI:iterID) (pI:calcID) tacis =
   let val (p, (ptp,_)) = get_state uI pI 
   in states:= 
      overwrite2 ((!states), ((uI, pI), (p, (ptp, tacis)))) end;*)
fun upd_tacis (cI: calcID) tacis = Synchronized.change states
        (fn s => case assoc (s, cI) of 
	          NONE => 
	            raise ERROR ("upd_tacis: calctree " ^ (string_of_int cI) ^ " not existent")
          | SOME ((ptp, _), us) => overwrite (s, (cI, ((ptp, tacis), us))));
(*///7.10.
fun upd_ipos (uI:iterID) (pI:calcID) (ip:pos') =
   let val (_, calc) = get_state uI pI 
   in states:= overwrite2 ((!states), ((uI, pI), (ip, calc))) end;*)
fun upd_ipos (cI: calcID) (uI: iterID) (ip: Pos.pos') = Synchronized.change states
        (fn s => case assoc (s, cI) of 
	          NONE =>
	            raise ERROR ("upd_ipos: calctree " ^ (string_of_int cI) ^ " not existent")
          | SOME _ => overwrite2 (s, ((cI, uI), ip)));


(** add and delete calcs **)

(*///7.10
fun add_pID (uI:iterID) (s:state) (p:(iterID * (calcID * state) list) list) = 
  let val new_ID = get_calcID uI p;
    val new_states = (the (assoc (p, uI))) @ [(new_ID, s)];
  in (new_ID, (overwrite (p, (uI, new_states)))) end;*)
(*
> val (new_calcID, new_states) = add_pID 1 (!states);
> states:= new_states;
> !states;
val it = [(3,[(#,#),(#,#)]),(4,[(#,#),(#,#)]),(1,[(#,#)])] : states
> val (new_calcID, new_states) = add_pID 3 (!states);
> states:= new_states;
> !states;
val it = [(3,[(#,#),(#,#),(#,#)]),(4,[(#,#),(#,#)]),(1,[(#,#)])] : states
> assoc2 (!states, (3, 1));
val it = SOME EmptyPtree : ctree option
> assoc2 (!states, (3, 2));
val it = NONE : ctree option
*)
(*///7.10
fun add_calc (uI:iterID) (s:state) = 
    let val (new_calcID, new_calcs) = add_pID uI s (!states)
    in states:= new_calcs; 
    new_calcID end; *)
fun add_user (cI: calcID) = Synchronized.change_result states
        (fn s => case assoc (s, cI) of 
	          NONE =>
	            raise ERROR ("add_user: calctree " ^ (string_of_int cI) ^ " not existent")
          | SOME (_, us) =>
	            let
                val new_uI = new_key us 1
	            in (new_uI: iterID, overwrite2 (s, ((cI, new_uI), Pos.e_pos'))) end);

(*///10.10.
fun del_calc (uI:iterID) (pI:calcID) = 
    (states:= del_assoc2 uI pI (!states); pI);*)
fun del_user (cI: calcID) (uI: iterID) =
      Synchronized.change_result states (fn s => (uI, del_assoc2 cI uI s));


(** add and delete calculations **)

(* here is the _only_ critical section on states,
  because a single calculation (with _one_ cI) is sequential *)
fun add_calc (cs: Calc.state_pre) = Synchronized.change_result states
        (fn s =>
            let val new_cI = new_key s 1
            in (new_cI: calcID, s @ [(new_cI, (cs, []))]) end);

(* delete doesn't report non existing elements *)
(*///7.10
fun del_user (uI:userID) = 
    (states:= del_assoc (!states, uI); uI);*)
fun del_calc (cI: calcID) = Synchronized.change_result states
        (fn s => (cI: calcID, del_assoc (s, cI)));

(* -------------- test all exported funs -------------- 
///7.10
Compiler.Control.Print.printDepth:=8;
states:=[];
add_user (); add_user (); !states;
ML> val it = 1 : userID
ML> val it = 2 : userID
ML> val it = [(1,[]),(2,[])]

val (hide,detail) = ([(["pI"],["tac"],true)]:hide,
		       [(["pI"],["tac"],true)]:detail);
add_calc 1 e_state; 
add_calc 1 (Calc.state_empty_pre,(hide,detail)); !states;
ML> val it = 1 : calcID
ML> val it = 2 : calcID
ML> val it =
  [(1,
    [(1,(((EmptyPtree,(#,#)),[]),([],[]))),
     (2,(((EmptyPtree,(#,#)),[]),([(#,#,#)],[(#,#,#)])))]),(2,[])]

val (pt,(p,p_)) = (EmptyPtree,e_pos');
val (pt,_) = cappend_problem pt p Uistate ([],References.empty);
upd_calc 1 2 ((pt,(p,p_)),[]); !states;
ML> val it =
  [(1,
    [(1,(((EmptyPtree,(#,#)),[]),([],[]))),
     (2,(((Nd #,(#,#)),[]),([(#,#,#)],[(#,#,#)])))]),(2,[])]
(*                          ~~~~~~~~~~~~~~~~~~~~ unchanged !!!*)

get_state 1 1; get_state 1 2;
ML> val it = (((EmptyPtree,([],Und)),[]),([],[])) : state
ML> val it =
  (((Nd
       (PblObj
          {branch=NoBranch,cell=[],env=(#,#,#,#),loc=(#,#),meth=[],
           model={Find=#,Given=#,Relate=#,Where=#,With=#},origin=(#,#),
           ostate=Incomplete,probl=[],result=(#,#),spec=(#,#,#)},[]),([],Und)),
    []),([(["pI"],["tac"],true)],[(["pI"],["tac"],true)])) : state

del_calc 2 1 (*non existent - NO msg!*); del_calc 1 2; !states;
ML> val it = [(1,[(1,(((EmptyPtree,(#,#)),[]),([],[])))]),(2,[])]

del_user 1; !states;
ML> val it = [(2,[])]

add_user (); add_user (); !states;
ML> val it = 1 : userID
ML> val it = 3 : userID
ML> val it = [(2,[]),(1,[]),(3,[])]
*)


(* -------------- test all exported funs -------------- 
print_depth 9;
states:=[];
add_calc Calc.state_empty_pre; add_calc Calc.state_empty_pre; !states;
|val it = 1 : calcID
|val it = 2 : calcID
|val it =
|   [(1, (((EmptyPtree, ([], Und)), []), [])),
|      (2, (((EmptyPtree, ([], Und)), []), []))]

add_user 2; add_user 2; !states; 
|val it = 1 : userID
|val it = 2 : userID
|val it =
|   [(1, (((EmptyPtree, ([], Und)), []), [])),
|      (2, (((EmptyPtree, ([], Und)), []), [(1, ([], Und)), (2, ([], Und))]))]


val cs = ((EmptyPtree, ([111], Und)), []) : Calc.state_pre;
upd_calc 1 cs; !states;
|val it =
|   [(1, (((EmptyPtree, ([111], Und)), []), [])),
|      (2, (((EmptyPtree, ([], Und)), []), [(1, ([], Und)), (2, ([], Und))]))]   

get_calc 1; get_calc 2;
|val it = ((EmptyPtree, ([111], Und)), []) : Calc.state_pre
|val it = ((EmptyPtree, ([], Und)), []) : Calc.state_pre

del_user 2 3 (*non existent - NO msg!*); del_user 2 1; !states;
|val it = 3 : userID
|val it = 1 : userID
|val it =
|   [(1, (((EmptyPtree, ([111], Und)), []), [])),
|      (2, (((EmptyPtree, ([], Und)), []), [(2, ([], Und))]))]

del_calc 1; !states;
|val it = 1 : calcID
|val it = [(2, (((EmptyPtree, ([], Und)), []), [(2, ([], Und))]))]

add_calc Calc.state_empty_pre; add_calc Calc.state_empty_pre; !states;
|val it = 1 : calcID
|val it = 3 : calcID
|val it =
|   [(2, (((EmptyPtree, ([], Und)), []), [(2, ([], Und))])),
|      (1, (((EmptyPtree, ([], Und)), []), [])),
|      (3, (((EmptyPtree, ([], Und)), []), []))]

add_user 2; !states;
|val it =
|   [(2, (((EmptyPtree, ([], Und)), []), [(2, ([], Und)), (1, ([], Und))])),
|      (1, (((EmptyPtree, ([], Und)), []), [])),
|      (3, (((EmptyPtree, ([], Und)), []), []))]
*)

(**)end(*struct*)
